; ***********************************
; ZX80 (8K ROM) / ZX81 Super Invasion
; ***********************************
; (c)1981 Beam Software
;
; RAM required: 1K (will not work with 4K or more)
;
; Disassembled by Paul Farrow, www.fruitcake.plus.com
; Initial version 24th April 2011


; =====
; Notes
; =====

; The machine code for the game is held in a numeric variable array. This means that RUN is not used to start the game. Instead it is started using GOTO 1.
; The game machine code bytes get copied from the variable array to memory beginning at $4000, and so the variable array also contains a small transfer routine.
; However, the BASIC program hard codes the address of this routine within the variables areas, and this will prevent the game running on a machine fitted
; with 4K RAM or more. This is because above 4K the display file is fully populated by the ZX81 ROM, thereby shifting the variables area up in memory.
; Ideally the BASIC should have looked up the address of the variables area from the VARS system variable and then applied a fixed offset.
;
; Display file is located at $4269. It consists of an initial NEWLINE followed by 16 rows, each ending with a NEWLINE. The first first row contains 16 columns
; where as all subsequent rows contain 24 columns. It is followed by 14 NEWLINE characters, i.e. empty rows. Not all of these are actually used to generate the
; display but provide a means to check whether an invader bomb has moved below the visible play area of the display file.
;
; The initial display file appears as follows:
;
; ...............
; ........................
; ......###..###..###..###
; .......#....#....#....#.
; ........................
; ......###..###..###..###
; .......#....#....#....#.
; ........................
; ......###..###..###..###
; .......#....#....#....#.
; ........................
; ........................
; ........................
; ........................
; ........................
; .#5#....................
;
; where '.' represents a space. The player's ship begins the game at row 14, column 1.
;
; The game cleverly implements the scrolling of the invaders by manipulating the number or blank rows before the visible play area. It removes all end of line
; terminators (NEWLINE characters) and replaces them with spaces. It then re-inserts them in either the column to the left or to the right. When scrolling the invaders
; down it replaces the first space within the visible display area with a NEWLINE thereby forcing a new blank line above the visible play area, i.e. the visible play
; area will get drawn one row lower down. End of row terminators (NEWLINE characters) are then inserted one character to the right of their previous positions.
;
; The game supports 3 speed levels, and so prompts for a numeric value before starting the game. Excepted values are 1 (fast), 2 (medium), 3 (slow).

; -------
; Markers
; -------
; The following markers appear throughout the disassembly:
;
; [...] = Indicates a comment about the code.
; ????  = Information to be determined.
;
; For bugs, the following marker format is used:
;
; [*BUG* - xxxx. Credit: yyyy]  = Indicates a confirmed bug, with a description 'xxxx' of it and the discoverer 'yyyy'.

; ---------------
; ZX80 Conversion
; ---------------
; Beam Software sold this game for both the ZX80 and ZX81. They achieved this by keeping the game code the same and using a custom video driver
; that generated the top and bottom borders in software, i.e. replicated the manner in which the ZX81 produces its display when in FAST mode.
; As a result, the same program could be used on the ZX80 fitted with the upgrade ROM. With only small changes to the video driver and some of
; the character set codes, the program could also be run on the ZX80 fitted with its original ROM. By comparing the differences used in other
; games available for both machines, it is possible to make a conjecture on how the ZX80 version of Super Invasion might have been.

; ---- ZX81 VERSION ----

#define CHAR_SET_ADDR           $1E00   ; The address of the space character bitmap.
#define CH_INVADER_BOTTOM       $03     ; Bottom character of an invader.
#define CH_INVADER_LEFT         $85     ; Left character of an invader.
#define CH_INVADER_MIDDLE       $03     ; Middle character of an invader.
#define CH_INVADER_RIGHT        $05     ; Right character of an invader.
#define CH_INVADER_BOMB         $04     ; Character for an invader bomb.
#define CH_PLAYER_SIDE          $83     ; Side character of the player's ship.

; ---- ZX80 VERSION ----
;
; #define CHAR_SET_ADDR         $0E00   ; The address of the space character bitmap.
; #define CH_INVADER_BOTTOM     $03     ; Bottom character of an invader.
; #define CH_INVADER_LEFT       $82     ; Left character of an invader.
; #define CH_INVADER_MIDDLE     $83     ; Middle character of an invader.
; #define CH_INVADER_RIGHT      $02     ; Right character of an invader.
; #define CH_INVADER_BOMB       $06     ; Character for an invader bomb.
; #define CH_PLAYER_SIDE        $03     ; Side character of the player's ship.


; ====================
; Assembler Directives
; ====================

; For use with the TASM assembler.

#define DEFB .BYTE
#define DEFW .WORD
#define DEFM .TEXT
#define DEFS .FILL
#define EQU  .EQU
#define ORG  .ORG
#define END  .END

; ======================================================================================================================================================

        ORG     $4009

; ================
; System Variables
; ================

VERSN:  DEFB $00                        ; $4009
E_PPC:  DEFW $03F2                      ; $400A
D_FILE: DEFW L40F3                      ; $400C
DF_CC:  DEFW L40F3+$0001                ; $400E
VARS:   DEFW L410C                      ; $4010
DEST:   DEFW $44C0                      ; $4012
E_LINE: DEFW L4388+$0001                ; $4014
CH_ADD: DEFW L4388+$000A                ; $4016
X_PTR:  DEFW $0000                      ; $4018
STKBOT: DEFW L4388+$000B                ; $401A
STKEND: DEFW L4388+$000B                ; $401C
BERG:   DEFB $00                        ; $401E
MEM:    DEFW $405D                      ; $401F
SPARE1: DEFB $00                        ; $4021
DF_SZ:  DEFB $02                        ; $4022
S_TOP:  DEFW $03F2                      ; $4023
LAST_K  DEFW $FDBF                      ; $4025
DBOUNC: DEFB $FF                        ; $4027
MARGIN: DEFB $37                        ; $4028
NXTLIN: DEFW L40F3                      ; $4029
OLDPPC: DEFW $0063                      ; $402B
FLAGX:  DEFB $00                        ; $402D
STRLEN: DEFW $CEE0                      ; $402E
T_ADDR: DEFW $0C8D                      ; $4030
SEED:   DEFW $0000                      ; $4032
FRAMES: DEFW $C3BC                      ; $4034
COORDS: DEFW $0000                      ; $4036
PR_CC:  DEFB $BC                        ; $4038
S_POSN: DEFB $21                        ; $4039
        DEFB $18                        ; $403A
CDFLAG: DEFB $00                        ; $403B
PRBUFF: DEFB $00, $00, $00, $00         ; $403C
        DEFB $00, $00, $00, $00
        DEFB $00, $00, $00, $00
        DEFB $00, $00, $00, $00
        DEFB $00, $00, $00, $00
        DEFB $00, $00, $00, $00
        DEFB $00, $00, $00, $00
        DEFB $00, $00, $00, $00
        DEFB $76                        ; $405C
MEMBOT: DEFB $00, $00, $00, $00         ; $405D
        DEFB $00, $00, $8A, $80
        DEFB $00, $00, $84, $A0
        DEFB $00, $00, $00, $80
        DEFB $80, $80, $80, $83
        DEFB $94, $96, $96, $91
        DEFB $90, $00, $00, $00
        DEFB $00, $00
SPARE2: DEFB $00                        ; $407B
SPARE3: DEFB $00                        ; $407C

; ======================================================================================================================================================

; =============
; BASIC Listing
; =============
;
; The BASIC listing is as follows:
;
;  100 PRINT "LEVEL?"
;  110 INPUT B
;  120 IF B<1 OR B>3 THEN GOTO 110
;  130 CLS
;  140 POKE 17178,30*B-26
; 1000 PRINT USR 17267
;
; [*BUG* - Line 1000 should call address 17269. Instead it executes Z80 instructions RLA followed by LD B,C
; before reaching the intended routine. Fortunately these instructions are harmless. Credit: Paul Farrow]

L407D:  DEFB    $00, $64                ; Line number 100.
        DEFW    $000A                   ; Line length.
        DEFB    $F5, $0B, $31, $2A      ; PRINT "LEVEL?"
        DEFB    $3B, $2A, $31, $0F
        DEFB    $0B
        DEFB    $76                     ; End of line.

        DEFB    $00, $6E                ; Line number 110.
        DEFW    $0003                   ; Line length.
        DEFB    $EE, $27                ; INPUT B
        DEFB    $76                     ; End of line.

        DEFB    $00, $78                ; Line number 120.
        DEFW    $0020                   ; Line length.
        DEFB    $FA, $27, $13, $1D      ; IF B<1 OR B>3 THEN GOTO 110
        DEFB    $7E, $81, $00, $00
        DEFB    $00, $00, $D9, $27
        DEFB    $12, $1F, $7E, $82
        DEFB    $40, $00, $00, $00
        DEFB    $DE, $EC, $1D, $1D
        DEFB    $1C, $7E, $87, $5C
        DEFB    $00, $00, $00
        DEFB    $76                     ; End of line.

        DEFB    $00, $82                ; Line number 130.
        DEFW    $0002                   ; Line length.
        DEFB    $FB                     ; CLS
        DEFB    $76                     ; End of line.

        DEFB    $00, $8C                ; Line number 140.
        DEFW    $0021                   ; Line length.
        DEFB    $F4, $1D, $23, $1D      ; POKE 17178,30*B-26
        DEFB    $23, $24, $7E, $8F
        DEFB    $06, $34, $00, $00
        DEFB    $1A, $1F, $1C, $7E
        DEFB    $85, $70, $00, $00
        DEFB    $00, $17, $27, $16
        DEFB    $1E, $22, $7E, $85
        DEFB    $50, $00, $00, $00
        DEFB    $76                     ; End of line.

        DEFB    $03, $E8                ; Line number 1000.
        DEFW    $000E                   ; Line length.
        DEFB    $F5, $D4, $1D, $23      ; PRINT USR 17267
        DEFB    $1E, $22, $23, $7E
        DEFB    $8F, $06, $E6, $00
        DEFB    $00
        DEFB    $76                     ; End of line.

; ======================================================================================================================================================

; ======================
; Collapsed Display File
; ======================

L40F3:  DEFB    $76
        DEFB    $76, $76, $76, $76
        DEFB    $76, $76, $76, $76
        DEFB    $76, $76, $76, $76
        DEFB    $76, $76, $76, $76
        DEFB    $76, $76, $76, $76
        DEFB    $76, $76, $76, $76

; ======================================================================================================================================================

; ==============
; Variables Area
; ==============

; Variable: A(126)

L410C:  DEFB    $86                     ; Array of numbers 'A'. [%100nnnnn, where nnnnn = letter - $20]
        DEFW    $0279                   ; Length of bytes that follow.
        DEFB    $01                     ; 1 dimension.
        DEFW    $007E                   ; Size of first dimention. This gives a payload area of 126*5=630 bytes.

; ------------------------------------------------------------------------------------------------------------------------------------------------------

; Variable: A(126) contents

        ORG     $4000

; =================
; Game Machine Code
; =================

        DEFB    $00                     ; Not used.

; =========================
; Video Generation Routines
; =========================

; -------------------------
; Delay Until Next TV Frame
; -------------------------

L4001:  DJNZ    L4001                   ; Execute a delay.

; -------------------
; Generate a TV Frame
; -------------------
; Exit: B'=Key press bitmap.

L4003:  OUT     ($FD),A                 ; Stop the vsync pulse.

; Generate the top border and the main display area.

; In FAST mode, the ZX81 uses software to generate the top and border borders, and this is what the routine below also does. In FAST mode, the ZX81 ROM takes 120 T-states
; from issuing OUT ($FD),A to before the CALL $02B5. The routine below takes 31 T-states, and hence is 89 T-states too short. The ROM loads the A register with $FC where as
; the routine below uses a value of $EC in an attempt to compensate for the deficit. The difference of $FC-$EC=$10 introduces 16*4=64 T-states, reducing the deficit to
; 89-64=25 T-states. If the A register is loaded with $E6 then this would introduce a 6*4=24 T-state delay, thereby reducing the deficit to 1 T-state.

        LD      A,$EC                   ; Timing constant for the first scan line of the top border. Set to $DD by the ROM for all subsequent scan lines.
        LD      B,$19                   ; There are 25 rows.
        LD      HL,$C269                ; The address of the display file with bit 15 set.
        LD      C,$30                   ; There are 48 lines in the top border. [Change this to $18 (24) scan lines for USA 60Hz]
        CALL    $02B5                   ; Output the top border and main display area.

; Generate the bottom border.

; The ZX81 ROM would take 105 T-states after returning from CALL $02B5 to before the next CALL $02B5. This routine takes 24 T-states, and
; hence is 81 T-states too short. The ZX81 ROM loads the A register with $FC and so this routine loads it with $E8 to compensate for the deficit.
; This difference of $FC-$E8=$14 introduces 20x4=80 T-states and hence reduces the deficit to -1 T-state.

        LD      A,$E8                   ; Timing constant for the first scan line of the bottom border. Set to $DD by the ROM for all subsequent scan lines.
        INC     B                       ; The bottom border is represented by one row.
        DEC     HL                      ; Point back to the last row's NEWLINE character, which is used now for the bottom border.
        LD      C,$2F                   ; There are 47 lines in the top border. [Change this to $17 (23) scan lines for USA 60Hz]
        CALL    $02B5                   ; Output the bottom border.

; ---- ZX80 VERSION ----
;
;       OUT     ($FF),A                 ; Stop the vsync pulse.
;
; Generate the top border and the main display area.
;
; The ZX80 ROM takes 38 T-states from issuing OUT ($FF),A to before the CALL $01AD. The routine below takes 31 T-states, and hence is 7 T-states too short.
; In addition, it calls $01B0 which omits the instruction LD C,(IY+$23) thereby reducing the time by a further 19 T-states. This makes the routine below
; 26 T-states too short. The ZX80 ROM loads the A register with $EC and so this routine loads it with $E6 to compensate for the deficit. This difference
; of $EC-$E6=$06 introduces 6x4=24 T-states and hence reduces the error to -2 T-states.
;
;       LD      A,$E6                   ; Timing constant for the first scan line of the top border. Set to $DD by the ROM for all subsequent scan lines.
;       LD      B,$19                   ; There are 25 rows.
;       LD      HL,$C269                ; The address of the display file with bit 15 set.
;       LD      C,$30                   ; There are 48 lines in the top border. [Change this to $18 (24) scan lines for USA 60Hz]
;       CALL    $01B0                   ; Output the top border and main display area.
;
; Generate the bottom border.
;
; The ZX80 ROM would take 52 T-states after returning from CALL $01AD to before the next CALL $01AD. This routine takes 24 T-states, and
; hence is 28 T-states too short. In addition, it calls $01B0 which omits the instruction LD C,(IY+$23) thereby reducing the time by a further
; 19 T-states. This makes the routine below 47 T-states too short. The ZX80 ROM loads the A register with $F3 and so this routine loads it
; with $E7 to compensate for the deficit. This difference of $F3-$E7=$0C introduces 12x4=48 T-states and hence reduces the error to +1 T-state.
;
;       LD      A,$E7                   ; Timing constant for the first scan line of the bottom border. Set to $DD by the ROM for all subsequent scan lines.
;       INC     B                       ; The bottom border is represented by one row.
;       DEC     HL                      ; Point back to the last row's NEWLINE character, which is used now for the bottom border.
;       LD      C,$2F                   ; There are 47 lines in the top border. [Change this to $17 (23) scan lines for USA 60Hz]
;       CALL    $01B0                   ; Output the bottom border.
;
; ----------------------

; Upper byte selects keyboard row to read.
;
;          Bit0  Bit1  Bit2  Bit3  Bit4    Bit4  Bit3  Bit2  Bit1  Bit0
;          ----  ----  ----  ----  ----    ----  ----  ----  ----  ----
; $F7FE =    1     2     3     4     5       6     7     8     9     0   = $EFFE
; $FBFE =    Q     W     E     R     T       Y     U     I     O     P   = $DFFE
; $FDFE =    A     S     D     F     G       H     J     K     L    N/L  = $BFFE
; $FEFE =  SHIFT   Z     X     C     V       B     N     M     .   SPACE = $7FFE

        LD      A,$E7                   ; Select the top row of the keyboard.
        IN      A,($FE)                 ; Read the keyboard.
        EXX
        LD      B,A                     ; Save the key press data in B'.
        EXX
        RET

; -----------------------------------
; Test if Below Invader can Drop Bomb
; -----------------------------------

L4022:  LD      BC,$0046                ; Offset to 3 rows down (from right side of current invader to left side of new invader).
        ADD     HL,BC                   ; Point to the invader position below.

        DEC     E                       ; Decrement the height counter.
        DEC     E
        DEC     E

; -----------------------------
; Test if Invader can Drop Bomb
; -----------------------------
; This routine searches the display file to determine whether an invader is present that could drop a bomb.
; Entry: E=Height of the bomb if aninvader found.
;        HL=Display file location to search.

L4029:  LD      A,CH_INVADER_BOTTOM     ; Graphic for the bottom of an invader.
        LD      BC,$0005                ; Search for up to 5 characters.
        CPIR                            ; Has the bottom of an invader been found?
        INC     C
        DEC     C                       ; Set the zero flag if C=$00, i.e. bottom of an invader not found.
        LD      B,$05                   ; Delay counter to balance paths.
        JR      Z,L4044                 ; Jump ahead if the bottom of an invader was not found.

; The bottom of an invader was found.

        LD      (L408C),HL              ; Save the display file address of the active invader bomb.
        PUSH    HL                      ; Save it.

        LD      HL,L408F                ; The store holding the display file location of the active invader bomb.
        LD      (HL),E                  ; Store the height of the invader bomb.

        POP     HL                      ; Fetch the display file address of the active invader bomb.
        LD      B,$00
        CPIR                            ; Perform remaining searches (which will fail) to ensure processing time takes a fixed length of time.

        INC     B                       ; Delay counter.

L4044:  DJNZ    L4044                   ; Delay to balance paths.

        RET

; ----------------------------------------------------
; Erase Player's Ship, Player Missile and Invader Bomb
; ----------------------------------------------------

L4047:  XOR     A			; A space.
        LD      HL,(L408C)              ; Fetch the display file address of the active invader bomb.
        LD      (HL),A                  ; Erase it.

        LD      HL,(L408A)              ; Fetch the display file address of the player missile.
        LD      (HL),A                  ; Erase it.

; -------------------
; Erase Player's Ship
; -------------------

L4050:  LD      HL,CHAR_SET_ADDR        ; Point to the character set bitmap for a space, i.e. three $00 in a row.

; ------------------
; Draw Player's Ship
; ------------------
; Entry: HL=Address of the graphic to draw.

L4053:  LD      DE,(L4082+$8000)        ; Fetch the display file address of the player. [*BUG* - Should be $4082, but will still work because of partial decoding of the RAM. Credit: Paul Farrow]
        LD      BC,$0003
        LDIR                            ; Copy 3 bytes of $00 to erase the player's ship.
        RET

; ----------------------------------------------------
; Shift Player's Ship, Player Missile and Invader Bomb
; ----------------------------------------------------
; Entry: DE=$FFFF (left) or $0001 (right), $FFE7 (up).

L405D:  LD      HL,(L408C)              ; Fetch the display file address of the active invader bomb.
        ADD     HL,DE                   ; Apply the left/right/up offset.
        LD      (L408C),HL              ; Save the new location.

        LD      HL,(L408A)              ; Fetch the display file address of the player missile.
        ADD     HL,DE                   ; Apply the left/right/up offset.
        LD      (L408A),HL              ; Save the new location.

        LD      HL,(L4082)              ; Fetch the display file location of the player.
        ADD     HL,DE                   ; Apply the left/right/up offset.
        LD      (L4082),HL              ; Save the new location.
        RET

; -----------------------------------
; Insert Display File Row Terminators
; -----------------------------------
; Entry: HL=Address in the display file of the top row's terminator position.

L4073:  LD      A,$76                   ; A NEWLINE character.

; -----------------------------------
; Change Display File Row Terminators
; -----------------------------------
; The routine is used to insert NEWLINE terminators for all rows, and also to remove them all.
; Entry: A=Character to place into the display file terminator locations.

L4075:  LD      B,$0F                   ; There are 15 rows.
        LD      DE,$0019                ; The offset to the next row.
        LD      (HL),A                  ; Insert the NEWLINE for space.

L407B:  ADD     HL,DE                   ; Advance to the next row.
        LD      (HL),A                  ; Insert the NEWLINE character.
        DJNZ    L407B                   ; Repeat for all rows.

        RET

; =================
; Program Variables
; =================

L4080:  DEFW    $0000                   ; Holds the display file address of the terminator of the 1st row.
L4082:  DEFW    $0000                   ; Holds the display file address of the start of the player's ship.
L4084:  DEFB    CH_PLAYER_SIDE          ; Graphic for the player.
        DEFB    $A1                     ; '5'.
        DEFB    CH_PLAYER_SIDE
L4087:  DEFB    $01                     ; Holds the direction of the invaders ($FF=Left, $01=Right).
L4088:  DEFB    $00                     ; Holds the invader movement distance counter. The invaders continuous to move in the current direction until the counter reaches $00.
L4089:  DEFB    $0C                     ; Holds a count of the number of invaders remaining.
L408A:  DEFW    $0000                   ; Holds the display file address of the player missile.
L408C:  DEFW    $0000                   ; Holds the display file address of the active invader bomb.
L408E:  DEFB    $00                     ; Holds a flag indictaing the distance an active player missile has to travel ($00=No missile).
L408F:  DEFB    $00                     ; Holds a flag indicating the distance an active invader bomb has to travel ($00=No bomb). Doubles as a leading space for the invader graphic.
L4090:  DEFB    CH_INVADER_LEFT         ; Graphic for the invader (including a trailing space).
        DEFB    CH_INVADER_MIDDLE
        DEFB    CH_INVADER_RIGHT
        DEFB    $00
L4094:  DEFB    $18                     ; Holds the invader movement timing counter. Subsequently set to $40 (slow), $22 (medium) or $04 (fast).

; ------------------------------------------------------------------------------------------------------------------------------------------------------

; ---------------
; Delay 4 Seconds
; ---------------
; The game has been lost or all invaders destroyed. Delay for 4 seconds before restarting.

L4095:  LD      D,$C8                   ; 4 seconds * 50 frames/sec.

; Enter a loop to generate TV frames.

L4097:  EXX                             ; Save the frame counter.

        LD      B,$B9                   ; Delay constant to pad out the vsync period.
        CALL    L4001                   ; Delay before generating a TV frame.

        EXX                             ; Retrieve the frame counter.
        DEC     D                       ; Decrement the frame counter.
        JR      NZ,L4097                ; Repeat for all frames.

; ===========
; Entry Point
; ===========

L40A1:  EXX

        LD      HL,$4269                ; The address of the display file.
        LD      SP,HL                   ; Point SP at the display file. This will make populating the display file simpler.

        EX      DE,HL
        PUSH    DE                      ; Save the address of the display file.

        LD      HL,$012D                ; This is an offset from the player's location to the highest invader immediately above the player.
        EXX                             ; Save it.

L40AC:  EX      AF,AF'
        LD      A,$0C                   ; AF'=Height of the highest possible invader bomb.
        EX      AF,AF'

        XOR     A                       ; The program bytes will be cleared and the display file populated using this.

        LD      HL,$43D9                ; The player begins at row 14, column 1.
        LD      (L4082),HL              ; Store the initial player position.

        LD      HL,L4084+$0001          ; Point to the graphic for the player.
        LD      (HL),$A1                ; Set the middle character to '5'.

        INC     HL
        INC     HL
        LD      (HL),$01                ; The initial direction of the invaders is to the right.
        INC     HL
        LD      (HL),$07                ; Set the the invader movement distance counter.
        INC     HL
        LD      (HL),$0C                ; There are 12 invaders to be destroyed.

; Clear program variables.

        LD      B,$06                   ; There are 6 bytes to clear to $00.

L40C8:  INC     HL
        LD      (HL),A                  ; Clear each program variable byte.
        DJNZ    L40C8                   ; Repeat for all program variable bytes.

; Fill the display file with spaces.

        POP     HL                      ; Fetch the address of the display file.
        PUSH    HL                      ; Save it again.

        LD      (HL),$76                ; Insert the NEWLINE for the top border.

        LD      B,$C8                   ; Insert 200 * 2 = 400 spaces.

L40D2:  INC     HL
        LD      (HL),A                  ; Insert a space.
        INC     HL
        LD      (HL),A                  ; Insert another space.
        DJNZ    L40D2                   ; Repeat for all 200 iterations.

; Insert all invaders into the display file.

        POP     HL                      ; Fetch the address of the display file.
        PUSH    HL                      ; Save it again.

        LD      C,$2F                   ; The offset to the first invader (top row of 16 + 1 row of 25 + 7 indent into row).
        ADD     HL,BC                   ; Generate the address of the first invader on the top row.
        PUSH    HL                      ; Save it.

        EX      DE,HL                   ; DE=Address of the top left invader position.
        LD      C,$05                   ; The are 5 characters per invader.
        LD      B,$04                   ; There are 4 invaders per row.

L40E3:  PUSH    BC                      ; Save the invader counter.

        LD      HL,L408F                ; The address of the graphic for an invader.
        LD      B,A
        LDIR

        POP     BC                      ; Retrieve the invader counter.
        DJNZ    L40E3                   ; Repeat for all invaders in this row.

        EX      DE,HL                   ; HL=Current display file address.
        LD      C,$07
        ADD     HL,BC                   ; Advance to the display file address of the middle of the left most invader on the next row.

        LD      DE,$0005                ; There is an offset of 5 to the next invader.
        LD      B,$04                   ; There are 4 invaders per row.
        LD      A,CH_INVADER_BOTTOM     ; The graphic for the bottom of the invader.

L40F8:  LD      (HL),A                  ; Insert the graphic into the display file.
        ADD     HL,DE                   ; Advance to the next invader location.
        DJNZ    L40F8                   ; Repeat for all invaders in this row.

        POP     HL                      ; Fetch the address of the first invader on the top row.
        PUSH    HL                      ; Save it again.

        LD      C,$4B                   ; The offset to the next row of invaders (3 rows of 25 columns).
        ADD     HL,BC
        EX      DE,HL                   ; DE=Address of the next row of invaders.

        POP     HL                      ; Fetch the address of the first invader on the top row.
        LD      C,$96
        LDIR                            ; Copy the display file bytes to generate a further 2 rows of invaders.

; Note the display file address of the first row's terminator.

        POP     HL                      ; Fetch the address of the display file.
        LD      C,$10                   ; The length of the first row.
        ADD     HL,BC                   ; Advance to the terminator position of the first row.
        LD      (L4080),HL              ; Save the display file address of the first row's terminator.

; Insert a NEWLINE after each row.

        CALL    L4073                   ; Insert display file row terminators.

; Insert 14 empty rows, i.e. just a NEWLINE terminator per row.
; This allow the invader bomb routine to detect when the bomb has reached the ground.

        LD      B,$0E

L4113:  INC     HL
        LD      (HL),A                  ; Insert a NEWLINE character.
        DJNZ    L4113                   ; Repeat for all empty rows.

; ------------------
; The Main Game Loop
; ------------------

L4117:  CALL    L4050                   ; Erase the player's ship.

        LD      HL,(L4082)              ; Fetch the display file address of the player.
        PUSH    HL
        POP     DE                      ; DE=Display file address of the player.

; Test for new key presses.

        EXX
        LD      A,B                     ; Fetch the key press bitmap.
        EXX

        BIT     2,A                     ; Was key 8 (or 3) pressed?
        JR      Z,L412C                 ; Jump if it was.

        BIT     4,A                     ; Was key 5 (or 6) pressed?
        JR      Z,L4132                 ; Jump if it was.

; No direction keys were being pressed.

        JR      L4131                   ; Jump ahead to continue.

; Key 8 (or 3) was pressed, i.e. the right key.

L412C:  BIT     4,A                     ; Was key 5 (or 6) also pressed?
        JR      Z,L4131                 ; Jump if it was.

; Only the right direction key was being pressed.

        INC     HL                      ; Advance to the next display file address.

; Joins here when no direction keys were being pressed.

L4131:  INC     HL                      ; Increment the display file address to counteract the decrement below.

; Joins here when the left direction key was being pressed.

L4132:  DEC     HL                      ; Left key: Move to the previous display file location. Right key: Counteract the increment above.
        PUSH    HL                      ; Save the new display file address.

        LD      A,(HL)                  ; Fetch the character at the new display file location, i.e. whether the left of the player will be.
        INC     HL
        INC     HL
        ADD     A,(HL)                  ; Add the character at the display file location where the right of the player will now be.

        POP     HL                      ; Retrieve the new display file location of the player.
        AND     A                       ; The player is currently erased so the result will be $00 if the player can move or $76 if not.
        JR      Z,L413D                 ; Jump ahead if the player can move.

; The player is at the edge of the display.

        EX      DE,HL                   ; Fetch the original display file address of the player.

L413D:  LD      (L4082),HL              ; Save the new display file address of the player.

; Draw the player's ship.

        LD      HL,L4084                ; The address of the graphic for the player.
        CALL    L4053                   ; Draw the player's ship in the display file.

; Check whether there is an active invader bomb to be moved.

        LD      HL,L408F                ; The address of the 'invader bomb active' flag.
        INC     (HL)
        DEC     (HL)                    ; Is there an active invader bomb?
        JR      Z,L4174                 ; Jump ahead if not.

; There is an active invader bomb.

        DEC     (HL)                    ; Decrement the height of the invader bomb.

        LD      HL,(L408C)              ; Fetch the display file address of the active invader bomb.
        LD      BC,$0019
        ADD     HL,BC                   ; Point to underneath it.
        LD      A,(HL)                  ; Fetch the character underneath the invader bomb.
        AND     A                       ; Has the bomb hit anything?
        JR      Z,L416B                 ; Jump if it has not hit anything.

; The player has been hit.

        XOR     A
        LD      (L408F),A               ; Clear the 'invader bomb active' flag.

        LD      HL,L4084+$0001          ; The address of the middle character of the player.
        DEC     (HL)                    ; Decrement the number of lives.
        LD      A,$9C                   ; Inverse '0'.
        CP      (HL)                    ; Have all lives been lost?
        JP      Z,L4095                 ; Jump to delay for 4 seconds before restarting the game.

; The player still has one or more lives remaining.

        LD      B,$7C                   ; Delay constant to pad out the remainder of the vsync period.
        JR      L418C                   ; Jump ahead to generate the new TV frame.

; An invader bomb can be moved.

L416B:  LD      (L408C),HL              ; Save the display file address of the invader bomb.
        LD      (HL),CH_INVADER_BOMB    ; Draw the invader bomb.

        LD      B,$7E                   ; Delay constant to pad out the remainder of the vsync period.
        JR      L418C                   ; Jump ahead to generate the new TV frame.

; There is no active invader bomb.

L4174:  LD      HL,(L4082)              ; Fetch the display file address of the player.

        EXX
        PUSH    HL                      ; Fetch the offset from the player's location to the highest invader immediately above the player.
        EXX
        POP     BC                      ; Transfer it to BC.

        AND     A
        SBC     HL,BC                   ; Point to the display file position on the highest invader row immediately above the player.

        EX      AF,AF'
        LD      E,A                     ; Fetch the height of the highest possible invader bomb.
        EX      AF,AF'

; Test the three levels of invaders to see if any of them remain and hence able to drop a bomb.
; If multiple invaders remain, then the lowest one takes precedence.

        CALL    L4029                   ; Test whether there is an invader on the top row that can drop a bomb.
        CALL    L4022                   ; Test whether there is an invader on the middle row that can drop a bomb.
        CALL    L4022                   ; Test whether there is an invader on the bottom row that can drop a bomb.

        LD      B,$46                   ; Delay constant to pad out the remainder of the vsync period.

; Joins here after the player has lost a life but has more remaining, and when an invader bomb is active and can be moved.

L418C:  CALL    L4001                   ; Delay before generating a TV frame.

; Test whether there is an active missile to be moved.

        LD      HL,L408E                ; Point to the active missile flag.
        LD      DE,$FFE8                ; DE holds a row offset of -24.
        INC     (HL)
        DEC     (HL)                    ; Is there an active missile?
        JR      Z,L41DC                 ; Jump if there is not.

; There is an active missile.

        DEC     (HL)                    ; Decrement the missile distance left counter.
        DEC     DE                      ; DE holds a row offset of -25, i.e. one whole row.

; Move the player's missile.

        LD      HL,(L408A)              ; Fetch the display file address of the player missile.
        ADD     HL,DE                   ; Point to the row above.
        LD      A,(HL)                  ; Fetch the character above the missle.
        AND     A                       ; Is the location occupied?
        JR      NZ,L41AC                ; Jump if it is.

; The player missile can be moved up.

        LD      (L408A),HL              ; Store the nrew display file address for the player missile.
        LD      (HL),$1B                ; Draw the player missile '.' at the new location.

        LD      B,$86                   ; Delay constant to pad out the duration of the vsync period.
        JR      L41F2                   ; Jump ahead to display the next TV frame.

; The player missile has hit something.

L41AC:  CP      CH_INVADER_BOMB         ; Is it an invader bomb?
        LD      A,$00                   ; This will be used to remove the active player missile.
        JR      NZ,L41B7                ; Jump if it is not.

; The player missile has hit an invader bomb.

        LD      (HL),A                  ; Erase the invader bomb.

        LD      B,$87                   ; Delay constant used to pad out the duration of the vsync period.
        JR      L41D7                   ; Jump ahead.

; The player missile has hit an invader.

L41B7:  DEC     DE                      ; Point back 2 characters to ensure the start of the invader is found.
        DEC     DE
        ADD     HL,DE                   ; Point to the row above, i.e. the first row of the invader.

; Erase the invader.

        LD      B,$03                   ; There are 3 rows to erase.

L41BC:  PUSH    BC                      ; Save the row counter.

        EX      DE,HL                   ; DE holds the display file address of the invader.
        LD      HL,CHAR_SET_ADDR        ; Point to the character set bitmap for a space, i.e. three $00 in a row.
        LD      BC,$0005                ; 5 bytes will be erased to ensure all of the invader is removed.
        LDIR                            ; Copy 5 bytes of $00 to erase the invader.

        EX      DE,HL
        LD      DE,$0014
        ADD     HL,DE                   ; Advance to the next row.

        POP     BC                      ; Retrieve the row counter.
        DJNZ    L41BC                   ; Repeat for each row of the invader.

; Test whether all invaders have been detroyed.

        LD      HL,L4089                ; Fetch the count of the number of invaders remaining.
        DEC     (HL)                    ; Decrement it.
        JP      Z,L4095                 ; Jump if all invaders have been destroyed to wait 4 seconds and then restart the game.

; Further invaders remain.

        LD      B,$56                   ; Delay constant to pad out the duration of the vsync period.

; Discard the current player missile.

L41D7:  LD      (L408E),A               ; Flag that there is not an active missile.
        JR      L41F2                   ; Jump ahead to display the next TV frame.

; There is not an active player missile.

L41DC:  LD      HL,(L4082)              ; Fetch the display file address of the player.
        ADD     HL,DE                   ; Add -24 to point to the location immediately above the centre of the player's ship.
        LD      (L408A),HL              ; Store it as the address of the missile.

        EXX
        LD      A,B                     ; Fetch the key press bitmap.
        EXX

        LD      B,$88                   ; Delay constant to pad out the duration of the vsync pulse.
        BIT     0,A                     ; Was the key 0 (or 1) pressed, i.e. the fire key?
        JR      NZ,L41F2                ; Jump ahead if it was not to display the next TV frame.

; The fire key was pressed.

        EX      AF,AF'
        LD      (L408E),A               ; Flag that there is an active missile by setting the missile distance counter.
        EX      AF,AF'

        DEC     B                       ; Adjust the delay constant.

L41F2:  DJNZ    L41F2                   ; Delay to pad out the duration of the vsync period.

; Generate a new TV frame.

        LD      HL,(L4080)              ; Fetch the display file address of the first row's terminator.
        CALL    L4073                   ; Insert display file row terminators.

        CALL    L4003                   ; Generate a TV frame.

; Now prepare to update all display elements.

        CALL    L4047                   ; Erase the player's ship, the player missile and invader bomb.

; Check whether it is time to move the invaders.

        LD      HL,L4094                ; Fetch the invader movement timing counter
        DEC     (HL)                    ; Decrement the counter. Is it time to move the invaders?
        JP      NZ,L4117                ; Jump back if not to perform another iteration of the main game loop.

; It is time to move the invaders.

        LD      (HL),$24                ; Reset the invader movement timing counter.
                                        ; The BASIC program will poke this (address 17178) with $04 (speed 1=fast), $22 (speed 2=medium), or $40 (speed 3=slow).

; Check whether it is time to reverse the direction of the invaders.

        LD      HL,L4088                ; Fetch the invader movement distance counter.
        DEC     (HL)                    ; Decrement it.
        JR      Z,L4232                 ; Jump if it is time to reverse the direction of the invaders.

; It is not time to reverse the invaders so move them in the current direction.

        LD      A,(L4087)               ; Fetch the direction of the invaders ($FF=Left, $01=Right).
        LD      E,A
        BIT     7,A                     ; Moving left or right?
        JR      NZ,L4218                ; Jump if moving to the right.

; Moving to the left.

        XOR     A                       ; Clear the high byte value of the offset.

L4218:  LD      D,A                     ; DE=$FFFF (left) or $0001 (right).
        CALL    L405D                   ; Apply the offset to display file locations of the player's ship, player missile and invader bomb.

        LD      HL,(L4080)              ; Fetch the display file address of the terminator of the 1st row.
        XOR     A                       ; The A register holds the code for a space.
        PUSH    HL
        PUSH    DE
        CALL    L4075                   ; Remove the current display file terminators, replacing them with spaces.
        POP     DE
        POP     HL
        ADD     HL,DE                   ; Apply the left/right offset to the first row terminator location.
        LD      (L4080),HL              ; Save the address.

        CALL    L4073                   ; Insert display file row terminators at their new locations.

        LD      B,$48                   ; Delay constant to pad out until the end of the vsync period.
        JR      L4255                   ; Jump ahead to display the next TV frame.

; Reverse the direction of the invaders.

L4232:  LD      (HL),$07                ; Reset the invader movement distance counter.

        LD      A,(L4087)               ; Fetch the invader movement direction.
        NEG                             ; Reverse it.
        LD      (L4087),A               ; Save the new direction.

; Move the invaders down a row.

        EXX
        LD      BC,$FFE7                ; The offset for a whole row, i.e. -25.
        ADD     HL,BC                   ; Adjust the offset from the player's location to the highest invader immediately above the player.
        EX      DE,HL                   ; Transfer to DE. HL=Address before the first visible character in the display file.

        INC     HL                      ; Point to the first visible character.
        LD      (HL),$76                ; Replace the character with a NEWLINE thereby causing the whole display to appear to shift downward.
        EX      DE,HL                   ; Save the new display file address in DE.

        PUSH    BC                      ; Transfer the row offset to DE.
        EXX
        POP     DE

        EX      AF,AF'
        DEC     A                       ; Decrement the missile distance counter.
        AND     A                       ; Has it reached 0, i.e. the invaders have landed?
        JP      Z,L4095                 ; Jump if it has to wiat 4 second and then restart the game.

        EX      AF,AF'                  ; Store the missile distance counter.

        CALL    L405D                   ; Shift the player's ship, player missile and invader bomb up a row.

        LD      B,$87                   ; Delay constant.

; Joins here when moving the invaders left or right.

L4255:  DJNZ    L4255                   ; Delay to pad out the duration of the vsync period.

        LD      HL,L4084                ; The address of the graphic for the player.
        CALL    L4053                   ; Draw the player's ship in the display file.

        CALL    L4003                   ; Generate a TV frame.
        JP      L4117                   ; Jump back to perform another iteration of the main game loop.

L4263:

; ======================================================================================================================================================

; ================
; Transfer Routine
; ================
; There is a bug in the BASIC program which actually runs the game from $4373.
; This causes the address in the jump instruction above to be interpreted as Z80
; instructions RLA followed by LD B,C. Fortunately these are harmless and so
; execution will drop through to here.

        ORG     $4375

L4375:  LD      HL,($4010)              ; Fetch the address of the VARS area.
        LD      BC,$0006                ;
        ADD     HL,BC                   ; Skip over the header size of the first variable.
        LD      DE,$4000                ; The destination address.
        LD      BC,$0280                ; The number of bytes to transfer [29 more than required].
        LDIR                            ; Transfer the bytes.
        JP      L40A1                   ; Jump to start the game.

        DEFB    $00                     ; Not used.

; ---- ZX80 VERSION ----
;
; The machine code is 611 bytes long. This is too large to fit within a single integer array on the
; ZX80 (which could hold 256 2-byte integers) and so it seems logical to span it into a second array.
; However, there is no evidence to suggest that this is how it was really done in the actual ZX80
; version of this game.
;
;       LD      HL,($4008)              ; Fetch the address of the VARS area.
;       INC     HL
;       INC     HL                      ; Skip over the header size of the first variable.
;       LD      DE,$4000                ; The destination address.
;       LD      BC,$0200                ; The number of bytes to transfer.
;       LDIR                            ; Transfer the bytes.
;
;       INC     HL
;       INC     HL                      ; Skip over the header size of the second variable.
;       LD      BC,$0063                ; The remaining number of bytes to transfer.
;       LDIR                            ; Transfer the bytes.
;       JP      L40A1                   ; Jump to start the game.
;
; ----------------------

; ======================================================================================================================================================

; =====================
; End of Variables Area
; =====================

L4388:  DEFB    $80

; ======================================================================================================================================================

; ==========
; FLOW CHART
; ==========

;                           $4117 +<------------------------------------------------------------------------------------+
;                                 |                                                                                     |
;                                 v                                                                                     |
;                         Erase Player's Ship                                                                           |
;                                 |                                                                                     |
;                                 v                                                                                     |
;                          Move Player's Ship                                                                           |
;                                 |                                                                                     |
;                                 v                                                                                     |
;                          Draw Player's Ship                                                                           |
;                                 |                                                                                     |
;                   No            v                                                                                     |
;         +<----------------- Active Bomb?                                                                              |
;         |                       |                                                                                     |
;         |                   Yes |                                                                                     |
;         |                       v                                                                                     |
;         |                   Move Bomb                                                                                 |
;         |                       |                                                                                     |
;         |                       v              No                                                                     |
;         |                  Hit Anything? -------------->+                                                             |
;         |                       |                       |                                                             |
;         |            Player Hit |                       |                                                             |
;         |                       v          No           v                                                             |
;         |                   More Lives? -------+    Draw Bomb                                                         |
;  Determine Whether              |              |        |                                                             |
;   An Invader Can            Yes |              v        |                                                             |
;    Drop A Bomb                  |           Restart     |                                                             |
;         |                       |                       |                                                             |
;         |                       v                       |                                                             |
;         +---------------------->+<----------------------+                                                             |
;                                 | $418C                                                                               |
;                                 v                                                                                     |
;                         Generate TV Frame                                                                             |
;                                 |                                                                                     |
;                   No            v                                                                                     |
;         +<-------------- Active Missile?                                                                              |
;         |                       |                                                                                     |
;         |                       v                                                                                     |
;         |                 Move Missile                                                                                |
;         |                       |                                                                                     |
;         |                       v          Yes                                                                        |
;         |                 Hit Anything? --------------->+                                                             |
;         |                       |                       |                                                             |
;         |                    No |                       v             No                                              |
;         |                       v               Hit Invader Bomb? --------------->+                                   |
;         |                 Draw Missile                  |                         |                                   |
;         |                       |                   Yes |                         v                                   |
;         |                       |                       v                   Erase Invader                             |
;         v           No          v                   Erase bomb                    |                                   |
;    Fire Pressed? -------------->+                       |                         v              Yes                  |
;         |                       |                       |              All Invaders Destroyed? -------+               |
;     Yes |                       |                       v                         |                   |               |
;         v                       |                 $41D7 +<------------------------+                   v               |
; Set Active Missile              |                       |                    No                    Restart            |
;         |                       v                       |                                                             |
;         +---------------------->+<----------------------+                                                             |
;                                 | $41F2                                                                               |
;                                 v                                                                                     |
;                         Generate TV Frame                                                                             | 
;                                 |                                                                                     |
;                                 v                                                                                     |
;                 Erase Player's Ship, Bomb, Missile                                                                    |
;                                 |                                                                                     |
;                                 v                                                                                     |
;                       Time To Move Invaders? ------------------------------------------------------------------------>+
;                                 |                                                                                     ^
;                                 v                                                                                     |
;                         Reverse Direction? ------------>+                                                             |
;                                 |                       |                                                             |
;                              No |                       v                                                             |
;                                 |             Toggle Direction Flag                                                   |
;                                 |                       |                                                             |
;                                 v                       v                                                             |
;                           Move Display             Move Display                                                       |
;                            Left/Right               Down A Row                                                        |
;                                 |                       |                                                             |
;                                 |                       v             Yes                                             |
;                                 |                Invaders Landed? ----------------+                                   |
;                                 |                       |                         |                                   |
;                                 v                    No |                         v                                   |
;                                 +<----------------------+                      Restart                                |
;                                 |                                                                                     |
;                                 v                                                                                     |
;                         Draw Player's Ship                                                                            |
;                                 |                                                                                     |
;                                 | $425D                                                                               |
;                                 v                                                                                     |
;                         Generate TV Frame                                                                             | 
;                                 |                                                                                     |
;                                 +-------------------------------------------------------------------------------------+

; ======================================================================================================================================================
                                 
                                         END
